"""
StateMachineLibrary is robot framework library to create state machine.
"""

from .interface import StateMachineLibrary

__author__ = "marcinooo"
__maintainer__ = "marcinooo"
__version__ = "1.0.1"


__all__ = [
    'StateMachineLibrary',
]



"""
Collections of exceptions raises by package.
"""


class StateMachineNotFoundError(Exception):
    """Error indicates that state machine was not found."""


class StateNotFoundError(Exception):
    """Error indicates that state was not added to state machine."""



"""
Library facade.
"""

from robot.api import logger

from .store import StateMachineStore
from .state_machine import StateMachine, State
from .utils import is_string, is_dictionary, get_keword, build_callback, dict_merge, keyword_should_exist
from .exceptions import StateMachineNotFoundError


class StateMachineFacade:
    """Facade provides api to create and manage state machine."""

    def __init__(self) -> None:
        self._store = StateMachineStore()

    def create_state_machine(self, name: str) -> None:
        """Creates state machine object."""
        if not is_string(name):
            raise RuntimeError('Name of state machine must be a string.')
        if self._store.get(name) is not None:
            logger.warn('State machine named {} has been overwritten'.format(name))
        sm = StateMachine(name)
        self._store.add(name, sm)
        logger.debug("State machine with name '{}' was created.".format(name))

    def add_state(self, state: str, on_update: str, sm: str) -> None:
        """Adds single state to state machine."""
        if not is_string(state):
            raise RuntimeError('State parameter should be name of keyword with main state procedure.')
        if not is_string(on_update):
            raise RuntimeError('On update parameter should be name of keyword with transition procedure.')
        if not is_string(sm):
            raise RuntimeError('Name of state machine must be a string.')
        sm_instance = self._get_state_machine_or_raise_error(sm)
        keyword_should_exist(state)
        keyword_should_exist(on_update)
        state_keyword = get_keword(state)
        on_update_keyword = get_keword(on_update)
        run_callback = build_callback(state_keyword)
        on_update_callback = build_callback(on_update_keyword)
        state_instance = State(name=state, run_callback=run_callback, on_update_callback=on_update_callback)
        sm_instance.add_state(state_instance)
        logger.debug("State with '{}' run keyword "
                     "and '{}' update keyword was add to '{}' state machine.".format(state, on_update, sm))

    def go_to_state(self, state: str, sm: str) -> None:
        """Jumps to specified state."""
        if not is_string(state):
            raise RuntimeError('State parameter should be name of keyword.')
        if not is_string(sm):
            raise RuntimeError('Sm parameter should be name of created state machine.')
        sm_instance = self._get_state_machine_or_raise_error(sm)
        sm_instance.go_to_state(state)

    def update_state(self, sm: str) -> None:
        """Goes to next state."""
        if not is_string(sm):
            raise RuntimeError('Sm parameter should be name of created state machine.')
        sm_instance = self._get_state_machine_or_raise_error(sm)
        sm_instance.update()

    def get_context(self, sm: str) -> dict:
        """Returns context for given state machine."""
        if not is_string(sm):
            raise RuntimeError('Sm parameter should be name of created state machine.')
        sm_instance = self._get_state_machine_or_raise_error(sm)
        return sm_instance.context

    def set_context(self, sm: str, context: dict) -> None:
        """Overwrites context for given state machine."""
        if not is_string(sm):
            raise RuntimeError('Sm parameter should be name of created state machine.')
        if not is_dictionary(context):
            raise RuntimeError('Context parameter should be a dictionary.')
        sm_instance = self._get_state_machine_or_raise_error(sm)
        sm_instance.context = context

    def update_context(self, sm: str, item: dict) -> None:
        """Updates context for given state machine."""
        if not is_string(sm):
            raise RuntimeError('Sm parameter should be name of created state machine.')
        if not is_dictionary(item):
            raise RuntimeError('Item parameter should be a dictionary.')
        sm_instance = self._get_state_machine_or_raise_error(sm)
        sm_instance.context = dict_merge(sm_instance.context, item)

    def destroy_state_machine(self, name: str) -> None:
        """Destroys state machine object."""
        if not is_string(name):
            raise RuntimeError('Name of state machine must be a string.')
        if self._store.get(name) is None:
            logger.warn("State machine named '{}' does not exist.".format(name))
        else:
            self._store.remove(name)
            logger.debug("State machine with name '{}' was destroyed.".format(name))

    def _get_state_machine_or_raise_error(self, sm: str) -> StateMachine:
        """
        Gets state machine with passed name or raises error if state machine does not exist.
        :param sm: name of state machine
        :return: state machine object
        """
        state_machine = self._store.get(sm)
        if state_machine is None:
            logger.debug('All created state machines:\n' + ',\n'.join(self._store.get_all()))
            missing_state_machine_message = "There is no state machine named '{}'.\n" \
                                            "Call keyword 'Create State Machine' to create it."
            raise StateMachineNotFoundError(missing_state_machine_message.format(sm))
        return state_machine



"""
Contains implementation of state machine with required utils.
"""

from typing import Callable
from .exceptions import StateNotFoundError
from .utils import log_state_machine_states


class State:
    """Creates state with required callbacks."""
    def __init__(self, name: str, run_callback: Callable, on_update_callback: Callable) -> None:
        self.name = name
        self.run_callback = run_callback
        self.on_update_callback = on_update_callback

    def run(self) -> None:
        """Runs main state logic."""
        self.run_callback()

    def update(self) -> None:
        """Runs transition logic."""
        self.on_update_callback()


class StateMachine:
    """Creates state machine."""
    def __init__(self, name: str) -> None:
        self.name = name
        self.current_state = None
        self.states = {}
        self.context = {}

    def add_state(self, state: State) -> None:
        """
        Adds single state to state machine.
        :param state: state to add
        :return: None
        """
        self.states[state.name] = state

    def go_to_state(self, state_name: str) -> None:
        """
        Jumps to specified state.
        :param state_name: name of next state
        :return: None
        """
        if state_name not in self.states:
            log_state_machine_states(self)
            raise StateNotFoundError("State '{}' was not added to '{}' state machine.".format(state_name, self.name))
        self.current_state = self.states[state_name]
        self.current_state.run()

    def update(self) -> None:
        """
        Goes to next state.
        :return: None
        """
        if self.current_state is not None:
            self.current_state.update()



"""
Library interface.
"""

from .facade import StateMachineFacade


class StateMachineLibrary:
    """Interface provides only necessary methods for robot framework."""

    def __init__(self) -> None:
        self._facade = StateMachineFacade()

    def create_state_machine(self, name) -> None:
        """
        Creates state machine object. State machine can control flow in your test or task definition.
        Example of flow:
            state A  ->  state B  ->  state C  ->  state A  ->  state B
                                              `->  state C
        :param name: name of state machine given by you
        :return: None
        """
        self._facade.create_state_machine(name)

    def add_state(self, state, on_update, sm) -> None:
        """
        Adds single state to state machine. To creates state you have to define state keyword
        which contains main code of state to execute. In addition you have to define a method that contains
        the transition logic to the next state.
        :param state: name of keyword with main state logic
        :param on_update: name of keyword with transition logic to the next state
        :param sm: name of state machine
        :return: None
        """
        self._facade.add_state(state, on_update, sm)

    def go_to_state(self, state, sm) -> None:
        """
        Jumps to specified state. State is identified by name of main keyword.

        .. warning::
            You can not use it in main keyword for given state.

        :param state: name of next keyword
        :param sm: name of state machine
        :return: None
        """
        self._facade.go_to_state(state, sm)

    def update_state(self, sm) -> None:
        """
        Goes to next state. It calls keyword with transition logic to the next state and main keyword from next state.
        Next state is indicated in keyword with transition logic for current state.
        :param sm: name of state machine
        :return: None
        """
        self._facade.update_state(sm)

    def get_context(self, sm: str) -> dict:
        """
        Returns context for given state machine. Context is dictionary where you can store common resources
        for state machine.
        :param sm: name of state machine
        :return: context as dictionary
        """
        return self._facade.get_context(sm)

    def set_context(self, sm: str, context: dict) -> None:
        """
        Overwrites context for given state machine. All common resources for state machine will be overwritten.
        :param sm: name of state machine
        :param context: context to set
        :return: None
        """
        self._facade.set_context(sm, context)

    def update_context(self, sm: str, item: dict) -> None:
        """
        Updates context for given state machine. It merges passed dictionary item with existing context.
        Example:
            +---------------------+------------------------------------+------------------------------------+
            | existing context    | item                               | result                             |
            +=====================+====================================+====================================+
            | {'result':{}}       | {'result':{'status':'FINISHED'}}   | {'result':{'status':'FINISHED'}}   |
            +---------------------+------------------------------------+------------------------------------+
            | {'time':'00:00'}    | {'user':'Leo'}                     | {'time':'00:00','user':'Leo'}      |
            +---------------------+------------------------------------+------------------------------------+
            | {'errors':0}        | {'errors':1}                       | {'errors':1}                       |
            +---------------------+------------------------------------+------------------------------------+

        :param sm: name of state machine
        :param item: object to add to context
        :return: None
        """
        self._facade.update_context(sm, item)

    def destroy_state_machine(self, name) -> None:
        """
        Destroys state machine object.
        :param name: name of created state machine
        :return: None
        """
        self._facade.destroy_state_machine(name)



"""
Contains different utils for whole package.
"""

import copy
from typing import Callable
from robot.api import logger
from robot.running.model import Keyword
from robot.running.context import EXECUTION_CONTEXTS
from robot.libraries.BuiltIn import RobotNotRunningError
from robot.errors import DataError
from robot.running.usererrorhandler import UserErrorHandler


def get_keword(name: str) -> Keyword:
    """
    Returns Keyword object for given name.
    :param name: name of keyword
    :return: found keyword
    """
    return Keyword(name)


def get_robot_context(top: bool = False) -> EXECUTION_CONTEXTS:
    """
    Returns current robot context or raises RobotNotRunningError error.
    :param top: top context will be returned if it is True
    :return: robot context
    """
    ctx = EXECUTION_CONTEXTS.current if not top else EXECUTION_CONTEXTS.top
    if ctx is None:
        raise RobotNotRunningError('Cannot access execution context')
    return ctx


def keyword_should_exist(name: str) -> None:
    """
    Checks if keyword was defined.
    :param name: name of keyword
    :return: None
    """
    ctx = get_robot_context()
    try:
        runner = ctx.namespace.get_runner(name)
    except DataError as error:
        raise AssertionError(error.message)
    if isinstance(runner, UserErrorHandler):
        raise AssertionError(runner.error.message)


def is_string(item: object) -> bool:
    """
    Checks if passed argument is a string object
    :param item: item for verify
    :return: Ture if item is string, otherwise False
    """
    return isinstance(item, str)


def is_dictionary(item: object) -> bool:
    """
    Checks if passed argument is a dictionary object
    :param item: item for verify
    :return: Ture if item is dictionary, otherwise False
    """
    return isinstance(item, dict)


def build_callback(keyword: Keyword) -> Callable:
    """
    Return function callback for state_machine.State class.
    :param keyword: keyword which 'run' method will be executed in callback
    :return: callback function
    """
    def callback():
        context = get_robot_context()
        keyword.run(context)
    return callback


def dict_merge(a: dict, b: dict) -> dict:
    """
    Merges second dictionary to first. It does not modify any of dictionaries.
    :param a: first dictionary
    :param b: second dictionary
    :return: merged dictionary
    """
    if not is_dictionary(a):
        raise ValueError("Argument 'a' should be a dictionary")
    if not is_dictionary(b):
        return b
    result = copy.deepcopy(a)
    for k, v in b.items():
        if k in result and isinstance(result[k], dict):
            result[k] = dict_merge(result[k], v)
        else:
            result[k] = copy.deepcopy(v)
    return result


def log_state_machine_states(sm) -> None:
    """
    Logs all created state machines.
    :param sm: State machine instnace
    :return: None
    """
    logger.debug('States added to state machine:\n' + ',\n'.join(sm.states.keys()))



import os
from setuptools import find_packages, setup


def read(rel_path: str) -> str:
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, rel_path)) as fp:
        return fp.read()


def get_version(rel_path: str) -> str:
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError('Unable to find version string.')


setup(
    name='robotframework-statemachinelibrary',
    version=get_version('src/StateMachineLibrary/__init__.py'),
    description='Robot Framework State Machine Library',
    long_description=read('README.rst'),
    author='Marcin Wachacki',
    license='MIT',
    url='https://github.com/marcinooo/robot-framework-state-machine/tree/main',
    package_dir={'': 'src'},
    packages=find_packages(
        where='src',
        exclude=['tests*'],
    ),
    install_requires=[
        'robotframework'
    ],
    py_modules=['StateMachineLibrary'],
)



"""
Contains implementation of store to collect all state machines.
"""

from .state_machine import StateMachine


class StateMachineStore:
    """Stores all state machines."""
    _STORE = {}
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(StateMachineStore, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def add(self, name: str, sm: StateMachine) -> None:
        """
        Adds state machine with given name to store.
        :param name: name for state machine
        :param sm: state machine object
        :return: None
        """
        self._STORE[name] = sm

    def get(self, name: str) -> StateMachine:
        """
        Returns state machine which is in store.
        :param name: name of state machine
        :return: state machine if it was found otherwise None
        """
        return self._STORE.get(name)

    def remove(self, name: str) -> None:
        """
        Removes state machine from store.
        :param name: name of state machine
        :return: None
        """
        try:
            del self._STORE[name]
        except KeyError:
            pass

    def get_all(self):
        """
        Returns all stored state machines.
        :return: dictionary with names and state machines
        """
        return self._STORE
